import os
import sys
from pathlib import Path
import gitignore_parser
import requests
import re
import uuid

host = "https://api.pycob.com" if os.environ.get("PYCOB_HOST") is None else os.environ.get("PYCOB_HOST")

build_id = str(uuid.uuid4())
print("----------------------------------------------------------------")
print(f"🌽 PyCob Deploy ---- ID: {build_id} 🌽")
print("----------------------------------------------------------------")

args = sys.argv[1:]

api_key = None
show_ui = True

current_directory = os.getcwd()
print(f"Current directory: {current_directory}")

if len(args) > 0:
    print("Checking for API key in args...")
    api_key = args[0]

    for arg in args:
        if arg == "--no-ui":
            # Check for requirements.txt in the current directory
            if not os.path.exists(os.path.join(current_directory, "requirements.txt")):
                print("🛑 Error: You must have a requirements.txt file to deploy without the UI.")
                sys.exit(1)
            if not os.path.exists(os.path.join(current_directory, "Procfile")):
                print("🛑 Error: You must have a Procfile file to deploy without the UI.")
                print("Example Procfile contents:")
                print("web: gunicorn --bind :$PORT --workers 1 --threads 8 --timeout 0 main:app")
                sys.exit(1)

            show_ui = False

# Check for API key
if api_key is None or api_key == "":
    # Check PYCOB_API_KEY environment variable
    print("Checking for API key in PYCOB_API_KEY environment variable...")
    api_key = os.environ.get("PYCOB_API_KEY")

    if api_key is None:
        if Path("pycob_api_key.txt").is_file():
            print("Checking for API key in pycob_api_key.txt...")
            f = open("pycob_api_key.txt", "r")
            api_key = f.read()

        if api_key is None:
            print("No API key found. ")
            print("Go to https://www.pycob.com to get an API key.")
            # Exit
            sys.exit(1)

def __send_api_request(endpoint: str, data: dict, api_key: str) -> dict:
    response = requests.post(host+"/rpc/"+endpoint, json=data, headers={"PYCOB-API-KEY": api_key})
    
    if response.status_code != 200:
        error_str = f"🛑 API request '{endpoint}' failed with status code " + str(response.status_code)
        print(error_str)
        return {"error": error_str}

    try:
        json_response = response.json()
        return json_response
    except:
        print("🛑 API request failed to return valid JSON.")
        return {"error": "API request failed to return valid JSON."}

def _url_for_file_storage(file_name):
    file = {
        "filename": file_name,
    }
    rv = __send_api_request("store_file", file, api_key)

    if 'url' in rv:
        return rv['url']
    else:
        print("🛑 Error: Unable to get URL for file storage")
        return ""

def store_code(file_path: str, build_id: str):
    remote_file_path = "code/" + build_id + file_path[1:]
    url = _url_for_file_storage(remote_file_path)

    if url == "":
        print("Error: Unable to store file")
        return

    # Put file with Content-Type: application/octet-stream
    file = open(file_path, 'rb')
    r = requests.put(url, data=file, headers={'Content-Type': 'application/octet-stream'})
    file.close()

    if r.status_code != 200:
        print("🛑 Error: Unable to store file")
        return
    else:
        print("✅")

# Recursively list all files in the current directory, excluding everything covered by .gitignore

# Get the list of ignored files
try:
    matches = gitignore_parser.parse_gitignore(".gitignore")
except:
    matches = None

# Any dotfile in any subdirectory is ignored
dotfile_pattern = re.compile(r'^\..*', re.IGNORECASE)


i = 0
max_files = 25


for root, dirs, files in os.walk("."):
    for file in files:
        path = os.path.join(root, file)
        if not path.startswith("./.") and not path.startswith("./venv/") and not path.startswith("./tmp/"):
            try:
                if matches and matches(os.path.join(root, file)):
                    continue
            
                if dotfile_pattern.match(file):
                    continue
            except:
                continue
            
            i = i + 1
            if i < max_files:
                print(f"📤 Uploading... File #{i}: {path}")
                store_code(path, build_id)
            elif i == max_files:
                print(f"Reached max file count. Skipping the rest...")

print("🍿 Done Uploading.")
server_response = __send_api_request("upload_complete", {"build_id": build_id}, api_key)                

if 'app_name' in server_response:
    app_name = server_response['app_name']

print("----------------------------------------------------------------")
print("🌽         Finish at https://www.pycob.com/deploy            🌽")
print("----------------------------------------------------------------")

import webbrowser
webbrowser.open("https://www.pycob.com/deploy?build_id="+build_id+"&app_name="+app_name)


from .request import Request
from .all_components import Page

def profile(server_request: Request) -> Page:
    page = Page("User Profile")

    username = server_request.get_username()

    if username is None or username == "":
        page.add_header("Not Logged In")
        page.add_link("Sign In", "/auth/login")
        return page

    card = page.add_card()

    card.add_header("User Profile")
    card.add_image("https://loremflickr.com/320/240", "User Profile Picture")
    card.add_header(username, size=2)

    form = card.add_form("/auth/__handle_logout", "POST")
    form.add_formsubmit("Sign Out")

    return page

# TODO: Put these functions inside closures that take a message parameter

def login_with_message(message: str):
    def login(server_request: Request) -> Page:
        redirect = server_request.get_query_parameter("redirect")

        page = Page("Sign In")

        card = page.add_card()

        card.add_header("Sign In")

        if message != "":
            card.add_alert(message, badge='Error', color='red')

        form = card.add_form("/auth/__handle_login", "POST")

        form.add_formtext("Username", "username", "username")
        form.add_formpassword()

        if redirect is not None and redirect != "":
            form.add_formhidden("redirect", redirect)

        form.add_formsubmit("Sign In")

        card.add_text("")
        card.add_link("Don't have an account? Sign Up", "/auth/signup")
        card.add_link("Forgot password?", "/auth/reset_password")

        return page
    
    return login

def signup(server_request: Request) -> Page:
    message = server_request.get_query_parameter("message")
    redirect = server_request.get_query_parameter("redirect")

    page = Page("Sign Up")    

    card = page.add_card()

    card.add_header("Sign Up")

    if message != "":
        card.add_alert(message, badge='Error', color='red')

    form = card.add_form("/auth/__handle_signup", "POST")

    form.add_formtext("New Username", "username", "username")
    form.add_formemail()
    form.add_formpassword()
    if redirect is not None and redirect != "":
        form.add_formhidden("redirect", redirect)

    form.add_formsubmit("Sign Up")

    card.add_text("")
    card.add_text("Already have an account?")
    card.add_link("Sign In", "/auth/login")

    return page

def logout(server_request: Request) -> Page:
    page = Page("Sign Out")

    page.add_header("You have been signed out.")

    page.add_link("Sign In", "/auth/login")

    return page

def reset_password(server_request: Request) -> Page:
    page = Page("Reset Password")

    card = page.add_card()

    card.add_header("Reset Password")

    form = card.add_form("/auth/__handle_reset_password", "POST")

    form.add_formtext("Your Username", "username", "username")
    form.add_formemail()

    form.add_formsubmit("Reset Password")

    card.add_text("")
    card.add_link("Sign In", "/auth/login")

    return page

def check_email(server_request: Request) -> Page:
    page = Page("Check Your Email")

    page.add_header("Thank you! Please check your e-mail for next steps.")

    return page

def change_password(server_request: Request) -> Page:
    page = Page("Change Password")

    return page

def forgot_password(server_request: Request) -> Page:
    page = Page("Forgot Password")

    return page

def update_email(server_request: Request) -> Page:
    page = Page("Update Email")

    return page

def verify_email(server_request: Request) -> Page:
    page = Page("Verify Email")

    return page

def verify_email_change(server_request: Request) -> Page:
    page = Page("Verify Email Change")

    return page




import flask
from .all_components import Page
from .request import Request
from .handler import PageHandler, LoginHandler, SignupHandler, LogoutHandler
import inspect
import os
import requests
from pathlib import Path
from .auth_forms import *
import sys
import string
import random
import pickle

demo_page = Page('Demo Page')
demo_page.add_header('Demo Page', 2)

class App:
    flask_app = None

    def __init__(self, name: str, subtitle="", app_nav=[], api_key=None, use_built_in_auth=False, profile_page=profile):
        self.flask_app = flask.Flask(__name__)
        self.name = name
        self.subtitle = subtitle
        self.nav = app_nav
        self.pages = {}
        self.api_key = api_key
        self.use_built_in_auth = use_built_in_auth
        self.profile_page = profile_page
        self.flask_app.add_url_rule('/favicon.ico', 'favicon.ico', redirect_to="https://cdn.pycob.com/favicon.ico")
        self.temp_dir = os.getcwd() + '/tmp/' + ''.join(random.choices(string.ascii_uppercase, k=5))
        self.error = None
        self.home_page_registered = False

        if use_built_in_auth:
            if self.api_key is None:
                self.__set_api_key()

            self.flask_app.secret_key = self.api_key
            self.__add_auth_pages()

    def __add_auth_pages(self):        
        self.__add_auth_page('/auth/login', login_with_message(""))
        self.__add_auth_page('/auth/login_retry', login_with_message("Unable to verify. Please try again."))
        self.__add_auth_page('/auth/signup', signup)
        self.__add_auth_page('/auth/logout', logout)
        self.__add_auth_page('/auth/reset_password', reset_password)
        self.add_page('/auth/profile', 'auth/profile', self.profile_page, show_in_navbar=False, footer_category=None, require_login=True)

        self.flask_app.add_url_rule('/auth/__handle_login', "__handle_login", view_func=LoginHandler(self), methods=["POST"])
        self.flask_app.add_url_rule('/auth/__handle_logout', "__handle_logout", view_func=LogoutHandler(self), methods=["POST"])
        self.flask_app.add_url_rule('/auth/__handle_signup', "__handle_signup", view_func=SignupHandler(self), methods=["POST"])


    def __check_login(self, redirect):
        if self.use_built_in_auth:
            if "username" not in flask.session:
                return flask.redirect("/auth/login?redirect=" + redirect)
        return None

    def __login(self, username: str):
        flask.session["username"] = username

    def __add_auth_page(self, url: str, page_function):
        self.flask_app.add_url_rule(url, url, view_func=PageHandler(self, page_function, redirect_url=None, protect_with_code=None), methods=["GET"])

    def send_email(self, from_email: str, to_email: str, subject: str, content: str):
        if self.api_key is None:
            self.__set_api_key()

        email = {
            "from_email": from_email,
            "to_email": to_email,
            "subject": subject,
            "content": content,
        }
        self.__send_api_request("send_email", email, self.api_key)

    def start_script(self, script_name: str, script_args: dict) -> str:
        if self.api_key is None:
            self.__set_api_key()

        script = {
            "script_name": script_name,
            "args": script_args,
        }
        rv = self.__send_api_request("run_job", script, self.api_key)

        if 'job_id' in rv:
            return rv['job_id']
        
        return ""

    def to_cloud_pickle(self, data, filename: str):
        # Turn data into a pickle with the given filename

        Path(self.temp_dir).mkdir(parents=True, exist_ok=True)

        file = open(self.temp_dir + "/" + filename, 'wb')    
        pickle.dump(data, file)
        file.close()

        url = self._url_for_file_storage(filename)

        if url == "":
            print("Error: Unable to store file")
            return
        
        # Put file with Content-Type: application/octet-stream
        file = open(self.temp_dir + "/" + filename, 'rb')
        requests.put(url, data=file, headers={'Content-Type': 'application/octet-stream'})
        file.close()

    def from_cloud_pickle(self, filename: str):
        try:
            with open(self.temp_dir + "/" + filename, 'rb') as existing_file:
                data = pickle.load(existing_file)
                existing_file.close()
                return data
        except:
            pass

        # Get the pickle with the given filename

        url = self._url_for_file_retrieval(filename)

        if url == "":
            print("Error: Unable to retrieve file")
            return

        # Get file with Accept: application/octet-stream
        rv = requests.get(url, headers={'Accept': 'application/octet-stream'})
        Path(self.temp_dir).mkdir(parents=True, exist_ok=True)
        file = open(self.temp_dir + "/" + filename, 'wb')
        file.write(rv.content)
        file.close()

        file = open(self.temp_dir + "/" + filename, 'rb')
        data = pickle.load(file)
        file.close()

        return data


    def _url_for_file_storage(self, file_name: str) -> str:
        if self.api_key is None:
            self.__set_api_key()

        file = {
            "filename": file_name,
        }
        rv = self.__send_api_request("store_file", file, self.api_key)

        if 'url' in rv:
            return rv['url']

        return ""
    
    def _url_for_file_retrieval(self, file_name: str) -> str:
        if self.api_key is None:
            self.__set_api_key()

        file = {
            "filename": file_name,
        }
        rv = self.__send_api_request("retrieve_file", file, self.api_key)

        if 'url' in rv:
            return rv['url']

        return ""

    def store_dict(self, table_id: str, object_id: str, value: dict):
        if self.api_key is None:
            self.__set_api_key()

        type_name = type(value).__name__
        print("Attempting to save a value of type = ", type_name)
        is_int = "int" in type_name
        is_float = "float" in type_name
        is_str = "str" in type_name
        is_list = "list" in type_name


        to_store = {"table_id": table_id, "object_id": object_id}

        if is_int or is_float or is_str:
            print("Saving flat value to DB")
            to_store['flat_value'] = value
        elif is_list:
            print("Saving list to DB")
            to_store['object_array'] = value
        else:
            print("Saving object to DB")
            to_store['object'] = value

        self.__send_api_request("store_object", to_store, self.api_key)

    def retrieve_dict(self, table_id: str, object_id: str) -> dict:
        if self.api_key is None:
            self.__set_api_key()

        server_response = self.__send_api_request("retrieve_object", {"object_id": object_id, "table_id": table_id}, self.api_key)

        if 'object' in server_response:
            return server_response['object']

    def delete_dict(self, table_id: str, object_id: str):
        if self.api_key is None:
            self.__set_api_key()

        self.__send_api_request("delete_object", {"object_id": object_id, "table_id": table_id}, self.api_key)        

    def query_dict(self, table_id: str, field_name: str, field_value) -> list:
        if self.api_key is None:
            self.__set_api_key()

        server_response = self.__send_api_request("query_objects", {"field_name": field_name, "table_id": table_id, "field_value": field_value}, self.api_key)

        if type(server_response) is list:
            arr = list(map(lambda x:  x['object'] if 'object' in x else None, server_response))
            num_items = len(arr)
            print(f"Got {num_items} items from the database table {table_id} where {field_name} == {field_value}")
            return arr

    def list_objects(self, table_id: str) -> list:
        if self.api_key is None:
            self.__set_api_key()

        server_response = self.__send_api_request("list_objects", {"table_id": table_id}, self.api_key)

        if type(server_response) is list:
            arr = list(map(lambda x:  x['object'] if 'object' in x else None, server_response))
            num_items = len(arr)
            print(f"Got {num_items} total items from the database table {table_id}")
            return arr

    def list_object_ids(self, table_id: str) -> list:
        if self.api_key is None:
            self.__set_api_key()

        server_response = self.__send_api_request("list_object_ids", {"table_id": table_id}, self.api_key)

        if server_response is not None and 'object_ids' in server_response:
            return server_response['object_ids']

    def store_secret(self, secret_name: str, secret_value: str):
        self.store_dict("_secrets", secret_name, {"value": secret_value})
    
    def retrieve_secret(self, secret_name: str) -> str:
        try:
            return self.retrieve_dict("_secrets", secret_name)['value']
        except:
            return None

    def get_quota_status(self) -> dict:
        if self.api_key is None:
            self.__set_api_key()

        server_response = self.__send_api_request("get_quota_status", {}, self.api_key)

        return server_response

    def generate_text_from_ai(self, prompt: str) -> str:
        if self.api_key is None:
            self.__set_api_key()

        server_response = self.__send_api_request("generate_text_from_ai", {"userPrompt": prompt}, self.api_key)

        if 'text' in server_response:
            return server_response['text']

    def generate_image_from_ai(self, prompt: str) -> str:
        if self.api_key is None:
            self.__set_api_key()

        server_response = self.__send_api_request("generate_image_from_ai", {"userPrompt": prompt}, self.api_key)

        if 'image' in server_response:
            return server_response['image']

    def __set_api_key(self):
        if self.api_key is None:
            # Check PYCOB_API_KEY environment variable
            self.api_key = os.environ.get("PYCOB_API_KEY")

            if self.api_key is None:

                if Path("pycob_api_key.txt").is_file():
                    # Check for api_key.txt file
                    f = open("pycob_api_key.txt", "r")
                    self.api_key = f.read()

                if self.api_key is None:
                    print("No API key found. ")
                    print("Go to https://www.pycob.com to get an API key.")
                    # Exit
                    self.error = 'No API key found. Go to <a href="https://www.pycob.com">https://www.pycob.com</a> to get an API key.'

    def __generate_api_key(self) -> dict:
        return self.__send_api_request("generate_key", {"app_name": self.name}, "")

    def __send_api_request(self, endpoint: str, data: dict, api_key: str) -> dict:
        response = requests.post("https://api.pycob.com/rpc/"+endpoint, json=data, headers={"PYCOB-API-KEY": api_key})
        
        if response.status_code != 200:
            error_str = "API request failed with status code " + str(response.status_code)
            print(error_str)
            return {"error": error_str}

        try:
            json_response = response.json()
            # print("API request returned: " + str(json_response))
            # print("Class of json_response: " + str(type(json_response)))
            return json_response
        except:
            print("API request failed to return valid JSON.")
            return {"error": "API request failed to return valid JSON."}

    def add_page(self, route: str, page_name: str, page_function, show_in_navbar=True, footer_category="All Pages", require_login=False, protect_with_code=None):
        print("Warning: add_page() is deprecated. Use register_function() instead.")
        endpoint_name = _strip_slashes(route)
        self.pages[endpoint_name] = {"page_name": page_name, "show_in_navbar": show_in_navbar, "footer_category": footer_category}

        redirect_url = None
        if require_login:
            if not self.use_built_in_auth:
                self.error = "You must set use_built_in_auth=True to use the require_login parameter."                
            redirect_url = "/"+endpoint_name

        self.flask_app.add_url_rule("/" + endpoint_name, endpoint_name, PageHandler(self, page_function, redirect_url, protect_with_code), methods=["GET", "POST"])

    def register_function(self, function, show_in_navbar=True, footer_category="All Pages", require_login=False, protect_with_code=None):
        endpoint_name = function.__name__
        self.pages[endpoint_name] = {"page_name": endpoint_name.replace("_", " ").title(), "show_in_navbar": show_in_navbar, "footer_category": footer_category}

        redirect_url = None
        if require_login:
            if not self.use_built_in_auth:
                self.error = "You must set use_built_in_auth=True to use the require_login parameter."                
            redirect_url = "/"+endpoint_name

        self.flask_app.add_url_rule("/" + endpoint_name, endpoint_name, PageHandler(self, function, redirect_url, protect_with_code), methods=["GET", "POST"])

        if not self.home_page_registered:
            self.home_page_registered = True
            self.flask_app.add_url_rule("/", "_home", PageHandler(self, function, redirect_url, protect_with_code), methods=["GET", "POST"])

    def run(self, port=8080, force_dev_mode=False):
        caller = inspect.currentframe().f_back
        called_from_module = caller.f_globals['__name__']
        print("Called from module ", called_from_module)
        if called_from_module == "__main__" or force_dev_mode:
            self.flask_app.run(debug=True, host='0.0.0.0', port=port)
            return None
        else:
            print(__name__)
            print("App is being run by a production server. Returning a runnable server object.")
            return self.flask_app


def _strip_slashes_from_pages_dict(pages_dict: dict) -> dict:
    new_pages_dict = {}
    for key, value in pages_dict.items():
        new_pages_dict[_strip_prefix_suffix(key, '/', '/')] = value
    return new_pages_dict

def _strip_slashes(text: str) -> str:
    return _strip_prefix_suffix(text, '/', '/')

def _strip_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix):]
    return text

def _strip_suffix(text: str, suffix: str) -> str:
    if text.endswith(suffix):
        return text[:-len(suffix)]
    return text

def _strip_prefix_suffix(text: str, prefix: str, suffix: str) -> str:
    return _strip_suffix(_strip_prefix(text, prefix), suffix)




from .component_interface import *


class AlertComponent(Component):
  def __init__(self, text: str, badge: str = '', color: str = 'indigo'):    
    self.text = text
    self.badge = badge
    self.color = color

  def to_html(self):
    return '''<div class="text-center py-4 lg:px-4">
<div class="p-2 bg-''' + self.color + '''-800 items-center text-''' + self.color + '''-100 leading-none lg:rounded-full flex lg:inline-flex" role="alert">
    <span class="flex rounded-full bg-''' + self.color + '''-500 uppercase px-2 py-1 text-xs font-bold mr-3">''' + self.badge + '''</span>
    <span class="font-semibold mr-2 text-left flex-auto">''' + self.text + '''</span>            
</div>
</div>'''

class CodeComponent(Component):
  def __init__(self, value: str, header: str = ''):    
    self.value = value
    self.header = header

  def to_html(self):
    return '''<div class="mx-auto my-10 max-w-3xl">
    <div class="flex h-11 w-full items-center justify-start space-x-1.5 rounded-t-lg bg-gray-900 px-3">
        <span class="h-3 w-3 rounded-full bg-red-400"></span>
        <span class="h-3 w-3 rounded-full bg-yellow-400"></span>
        <span class="h-3 w-3 rounded-full bg-green-400"></span>
        <code class="pl-5 text-lime-500">''' + self.header + '''</code>
    </div>
    <div class="w-full border-t-0 bg-gray-700 pb-5">
        <code class="text-gray-500">&gt&gt&gt</code>
        <code class="text-white">''' + self.value + '''</code>
    </div>
</div>'''

class CodeeditorComponent(Component):
  def __init__(self, value: str, language: str = 'python'):    
    self.value = value
    self.language = language

  def to_html(self):
    return '''<style type="text/css" media="screen">
#editorContainer {
    width: calc( 100vw - 40px );
    height: 500px;
    max-height: calc( 80vh - 60px );
    position: relative;
    background-color: red;
}
#editor { 
    position: absolute;
    top: 0;
    right: 0;
    bottom: 0;
    left: 0;
}
</style>

<div id="editorContainer">
    <div id="editor">''' + self.value + '''</div> 
</div>
<script src="https://cdn.jsdelivr.net/gh/ajaxorg/ace-builds/src-noconflict/ace.js" type="text/javascript" charset="utf-8"></script>
<script>
    var editor = ace.edit("editor");
    editor.setTheme("ace/theme/monokai");
    editor.session.setMode("ace/mode/''' + self.language + '''");

    const savedCode = localStorage.getItem('code');

    if (savedCode) {
        editor.setValue(savedCode);
    }
</script>'''

class DividerComponent(Component):
  def __init__(self):    
    pass

  def to_html(self):
    return '''<hr class="my-5 border-gray-300 w-full">'''

class FooterComponent(Component):
  def __init__(self, title: str, subtitle: str = '', logo: str = '', components: list = None):    
    self.title = title
    self.subtitle = subtitle
    self.logo = logo
    # https://stackoverflow.com/questions/4841782/python-constructor-and-default-value
    self.components = components or []

  def to_html(self):
    return '''<footer class="text-gray-600 body-font">
    <div class="container px-5 py-24 mx-auto flex md:items-center lg:items-start md:flex-row md:flex-nowrap flex-wrap flex-col">
        <div class="w-64 flex-shrink-0 md:mx-0 mx-auto text-center md:text-left">
            <a class="flex title-font font-medium items-center md:justify-start justify-center text-gray-900"><img class="object-scale-down h-10" src="''' + self.logo + '''"><span class="ml-3 text-xl">''' + self.title + '''</span></a>
            <p class="mt-2 text-sm text-gray-500">''' + self.subtitle + '''</p>
        </div>
        <div class="flex-grow flex flex-wrap md:pl-20 -mb-10 md:mt-0 mt-10 md:text-left text-center">
            ''' + '\n'.join(map(lambda x: x.to_html(), self.components)) + ''' 
        </div>
    </div>
</footer>'''

  def add(self, component):
    self.components.append(component)
    return self

  def add_component(self, component):
    self.components.append(component)
    return self
  def add_footercategory(self, title: str, components: list = None):    
    self.components.append(FootercategoryComponent(title, components))
    return self
    

class FootercategoryComponent(Component):
  def __init__(self, title: str, components: list = None):    
    self.title = title
    # https://stackoverflow.com/questions/4841782/python-constructor-and-default-value
    self.components = components or []

  def to_html(self):
    return '''<div class="lg:w-1/4 md:w-1/2 w-full px-4">
    <h2 class="title-font font-medium text-gray-900 tracking-widest text-sm mb-3 uppercase">''' + self.title + '''</h2>
    <nav class="list-none mb-10">
        ''' + '\n'.join(map(lambda x: x.to_html(), self.components)) + ''' 
    </nav>
</div>'''

  def add(self, component):
    self.components.append(component)
    return self

  def add_component(self, component):
    self.components.append(component)
    return self
  def add_footerlink(self, title: str, url: str):    
    self.components.append(FooterlinkComponent(title, url))
    return self
    

class FooterlinkComponent(Component):
  def __init__(self, title: str, url: str):    
    self.title = title
    self.url = url

  def to_html(self):
    return '''<li><a href="''' + self.url + '''" class="text-gray-600 hover:text-gray-800">''' + self.title + '''</a></li>'''

class FormComponent(Component):
  def __init__(self, action: str, components: list = None, method: str = 'GET'):    
    self.action = action
    # https://stackoverflow.com/questions/4841782/python-constructor-and-default-value
    self.components = components or []
    print("Form components on init = ", self.components)
    self.method = method

  def to_html(self):
    return '''<form action="''' + self.action + '''" method="''' + self.method + '''">
    ''' + '\n'.join(map(lambda x: x.to_html(), self.components)) + ''' 
</form>'''

  def add(self, component):
    self.components.append(component)
    return self

  def add_component(self, component):
    self.components.append(component)
    return self
  def add_formtext(self, label: str, name: str, value: str):    
    print("Form components on add_formtext = ", self.components)
    self.components.append(FormtextComponent(label, name, value))
    return self
    

  def add_formsubmit(self, label: str = 'Submit'):    
    self.components.append(FormsubmitComponent(label))
    return self
    

class FormsubmitComponent(Component):
  def __init__(self, label: str = 'Submit'):    
    self.label = label

  def to_html(self):
    return '''<button type="submit" class="text-white bg-indigo-500 border-0 py-2 px-6 focus:outline-none hover:bg-indigo-600 rounded text-lg">''' + self.label + '''</button>'''

class FormtextComponent(Component):
  def __init__(self, label: str, name: str, value: str):    
    self.label = label
    self.name = name
    self.value = value

  def to_html(self):
    return '''<div class="relative mb-4">
    <label for="''' + self.name + '''" class="leading-7 text-sm text-gray-600">''' + self.label + '''</label>
    <input name="''' + self.name + '''" type="text" class="w-full bg-white rounded border border-gray-300 focus:border-indigo-500 focus:ring-2 focus:ring-indigo-200 text-base outline-none text-gray-700 py-1 px-3 leading-8 transition-colors duration-200 ease-in-out" id="''' + self.name + '''" value="''' + self.value + '''">
</div>'''

class HeaderComponent(Component):
  def __init__(self, text: str, size: int = 1):    
    self.text = text
    self.size = size

  def to_html(self):
    return '''<p class="title-font sm:text-''' + str(self.size) + '''xl text-xl font-medium text-gray-900 mb-3">''' + self.text + '''</p>'''

class HeroComponent(Component):
  def __init__(self, title: str, subtitle: str = '', image: str = '', color: str = 'indigo'):    
    self.title = title
    self.subtitle = subtitle
    self.image = image
    self.color = color

  def to_html(self):
    return '''<section class="bg-white">
    <div class="grid max-w-screen-xl px-4 py-8 mx-auto lg:gap-8 xl:gap-0 lg:py-16 lg:grid-cols-12">
        <div class="mr-auto place-self-center lg:col-span-7">
            <h1 class="max-w-2xl mb-4 text-4xl font-extrabold tracking-tight leading-none md:text-5xl xl:text-6xl">''' + self.title + '''</h1>
            <p class="max-w-2xl mb-6 font-light text-gray-500 lg:mb-8 md:text-lg lg:text-xl">''' + self.subtitle + '''</p>
            <a href="#" class="inline-flex items-center justify-center px-5 py-3 mr-3 text-base font-medium text-center text-white rounded-lg bg-''' + self.color + '''-700 hover:bg-''' + self.color + '''-800 focus:ring-4 focus:ring-''' + self.color + '''-300">
                Button 1                        
            </a>
            <a href="#" class="inline-flex items-center justify-center px-5 py-3 text-base font-medium text-center text-gray-900 border border-gray-300 rounded-lg hover:bg-gray-100 focus:ring-4 focus:ring-gray-100">
                Button 2
            </a> 
        </div>
        <div class="hidden lg:mt-0 lg:col-span-5 lg:flex">
            <img src="''' + self.image + '''" >
        </div>                
    </div>
</section>'''

class HtmlComponent(Component):
  def __init__(self, value: str):    
    self.value = value

  def to_html(self):
    return '''''' + self.value + ''''''

class ImageComponent(Component):
  def __init__(self, url: str, alt: str):    
    self.url = url
    self.alt = alt

  def to_html(self):
    return '''<img class="max-w-fit" src="''' + self.url + '''" alt="''' + self.alt + '''">'''

class LinkComponent(Component):
  def __init__(self, text: str, url: str, classes: str = ''):    
    self.text = text
    self.url = url
    self.classes = classes

  def to_html(self):
    return '''<a class="''' + self.classes + '''" href="''' + self.url + '''">''' + self.text + '''</a>'''

class NavbarComponent(Component):
  def __init__(self, title: str, logo: str = '', components: list = None):    
    self.title = title
    self.logo = logo
    # https://stackoverflow.com/questions/4841782/python-constructor-and-default-value
    self.components = components or []

  def to_html(self):
    return '''<header class="text-white body-font">
    <div class="gradient-background mx-auto flex flex-wrap p-5 flex-col md:flex-row items-center">
        <a class="flex title-font font-bold items-center text-gray-100 mb-4 md:mb-0"><img class="object-scale-down h-10" src="''' + self.logo + '''"><span class="ml-3 text-4xl">''' + self.title + '''</span></a>
        <nav class="md:ml-auto flex flex-wrap items-center text-base justify-center">
          <button onclick="toggleDarkMode()" type="button" class="mx-3 px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800 bg-gradient-to-br from-purple-600 to-blue-500 group-hover:from-purple-600 group-hover:to-blue-500">
            Dark Mode
          </button>            
            ''' + '\n'.join(map(lambda x: x.to_html(), self.components)) + ''' 
        </nav>
        <button class="inline-flex items-center bg-gray-100 text-black border-0 py-1 px-3 focus:outline-none hover:bg-gray-200 rounded text-base mt-4 md:mt-0">Sign In<svg fill="none" stroke="currentColor" stroke-linecap="round" stroke-linejoin="round" stroke-width="2" class="w-4 h-4 ml-1" viewBox="0 0 24 24"><path d="M5 12h14M12 5l7 7-7 7"></path></svg></button>
    </div>
</header>'''

  def add(self, component):
    self.components.append(component)
    return self

  def add_component(self, component):
    self.components.append(component)
    return self
  def add_link(self, text: str, url: str, classes: str = ''):    
    self.components.append(LinkComponent(text, url, classes))
    return self
    

class Page(Component):
  def __init__(self, title: str, components: list = None, auto_navbar: bool = True, auto_footer: bool = True):    
    self.title = title
    # https://stackoverflow.com/questions/4841782/python-constructor-and-default-value
    self.components = components or []
    self.auto_navbar = auto_navbar
    self.auto_footer = auto_footer

  def to_html(self):
    return '''<div class="container px-5 py-24 mx-auto max-w-fit">
    ''' + '\n'.join(map(lambda x: x.to_html(), self.components)) + ''' 
</div>'''

  def add(self, component):
    self.components.append(component)
    return self

  def add_component(self, component):
    self.components.append(component)
    return self
  def add_html(self, value: str):    
    self.components.append(HtmlComponent(value))
    return self
    

  def add_text(self, value: str):    
    self.components.append(TextComponent(value))
    return self
    

  def add_link(self, text: str, url: str, classes: str = ''):    
    self.components.append(LinkComponent(text, url, classes))
    return self
    

  def add_image(self, url: str, alt: str):    
    self.components.append(ImageComponent(url, alt))
    return self
    

  def add_header(self, text: str, size: int = 1):    
    self.components.append(HeaderComponent(text, size))
    return self
    

  def add_alert(self, text: str, badge: str = '', color: str = 'indigo'):    
    self.components.append(AlertComponent(text, badge, color))
    return self
    

  def add_hero(self, title: str, subtitle: str = '', image: str = '', color: str = 'indigo'):    
    self.components.append(HeroComponent(title, subtitle, image, color))
    return self
    

  def add_code(self, value: str, header: str = ''):    
    self.components.append(CodeComponent(value, header))
    return self
    

  def add_divider(self):    
    self.components.append(DividerComponent())
    return self
    

  def add_section(self, id: str, name: str, level: int = 1):    
    self.components.append(SectionComponent(id, name, level))
    return self
    

  def add_form(self, action: str, components: list = None, method: str = 'GET'):    
    self.components.append(FormComponent(action, components, method))
    return self
    


  def add_pandastable(self, dataframe: str):
    advanced_add_pandastable(self, dataframe)
    return self
    

  def add_navbar(self, title: str, logo: str = '', components: list = None):    
    self.components.append(NavbarComponent(title, logo, components))
    return self
    

  def add_footer(self, title: str, subtitle: str = '', logo: str = '', components: list = None):    
    self.components.append(FooterComponent(title, subtitle, logo, components))
    return self
    

  def add_sidebar(self, components: list = None):    
    self.components.append(SidebarComponent(components))
    return self
    

  def add_codeeditor(self, value: str, language: str = 'python'):    
    self.components.append(CodeeditorComponent(value, language))
    return self
    

class SectionComponent(Component):
  def __init__(self, id: str, name: str, level: int = 1):    
    self.id = id
    self.name = name
    self.level = level

  def to_html(self):
    return '''<span id=''' + self.id + '''></span>'''

class SidebarComponent(Component):
  def __init__(self, components: list = None):    
    # https://stackoverflow.com/questions/4841782/python-constructor-and-default-value
    self.components = components or []

  def to_html(self):
    return '''<aside class="hidden md:block flex w-72 flex-col space-y-2 bg-slate-200 p-2 h-screen sticky top-0">
    <div class="sticky top-0">
        ''' + '\n'.join(map(lambda x: x.to_html(), self.components)) + ''' 
    </div>
</aside>'''

  def add(self, component):
    self.components.append(component)
    return self

  def add_component(self, component):
    self.components.append(component)
    return self
  def add_sidebarcategory(self, title: str, components: list = None):    
    self.components.append(SidebarcategoryComponent(title, components))
    return self
    

class SidebarcategoryComponent(Component):
  def __init__(self, title: str, components: list = None):    
    self.title = title
    # https://stackoverflow.com/questions/4841782/python-constructor-and-default-value
    self.components = components or []

  def to_html(self):
    return '''<div class="mb-8">
    <h2 class="text-lg font-medium text-gray-900 tracking-wider uppercase mb-3">''' + self.title + '''</h2>
    <ul class="ml-5 list-none">
        ''' + '\n'.join(map(lambda x: x.to_html(), self.components)) + ''' 
    </ul>
</div>'''

  def add(self, component):
    self.components.append(component)
    return self

  def add_component(self, component):
    self.components.append(component)
    return self
  def add_sidebarlink(self, title: str, url: str):    
    self.components.append(SidebarlinkComponent(title, url))
    return self
    

class SidebarlinkComponent(Component):
  def __init__(self, title: str, url: str):    
    self.title = title
    self.url = url

  def to_html(self):
    return '''<li><a href="''' + self.url + '''" class="text-gray-600 hover:text-gray-800">''' + self.title + '''</a></li>'''

class TextComponent(Component):
  def __init__(self, value: str):    
    self.value = value

  def to_html(self):
    return '''<p>''' + self.value + '''</p>'''




from urllib.parse import quote
import re
import json

def advanced_add_pandastable(self, df, hide_fields, action_buttons):
    cols_to_show = []
    show_actions = False

    if action_buttons is None:
        action_buttons = []

    if len(__get_action_buttons_to_add(action_buttons)) > 0:
        show_actions = True

    for col in df.columns:
        if col not in hide_fields:
            cols_to_show.append(col)

    # Pandas dataframe to html
    html = '''<div class="p-8">'''

    html += '''<div class="relative overflow-x-auto shadow-md sm:rounded-lg">'''

    html += '''<table class="w-full text-sm text-left text-gray-500 dark:text-gray-400">'''

    # Pandas DataFrame columns to 
    html += '''<thead class="text-xs text-gray-700 uppercase bg-gray-50 dark:bg-gray-700 dark:text-gray-400">'''

    html += "<tr>"

    # Get df index name
    if df.index.name is not None:
        html += '''<th scope="col" class="px-6 py-3">''' + df.index.name +  "</th>"

    if show_actions:
        html += '''<th scope="col" class="px-6 py-3">Actions</th>'''

    for column in df.columns:
        if column in cols_to_show:
            html += '''<th scope="col" class="px-6 py-3">''' + column + "</th>"

    html += "</tr>"

    html += "</thead>"

    # Pandas DataFrame rows to html
    html += "<tbody>"
    i = 0
    for index, row in df.iterrows():
        record = row.to_dict()

        if i % 2 == 0:
            html += '''<tr class="bg-white border-b dark:bg-gray-900 dark:border-gray-700">'''
        else:
            html += '''<tr class="bg-gray-50 border-b dark:bg-gray-800 dark:border-gray-700">'''

        if show_actions:
            html += '''<td class="px-6 py-4">'''
        
            action_buttons_to_add = __get_action_buttons_to_add(action_buttons)
            
            for button_to_add in action_buttons_to_add:
                hydrated_label = button_to_add.label.format(**record)
                hydrated_url = button_to_add.url.format(**record)

                if button_to_add.open_in_new_window:
                    html += """<a href='""" + hydrated_url + """' target="_blank" class="px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800 whitespace-nowrap mr-1">""" + hydrated_label + """</button>"""
                else:
                    html += """<a href='""" + hydrated_url + """' class="px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800 whitespace-nowrap mr-1">""" + hydrated_label + """</button>"""


            html += '''</td>'''

        for column in df.columns:
            key = column
            if column in cols_to_show:
                action_button = __find_key_in_action_buttons(key, action_buttons)

                value = __format_python_object_for_json(record[key])

                if action_button is not None:
                    hydrated_label = action_button.label.format(**record)
                    hydrated_url = action_button.url.format(**record)
                    if action_button.open_in_new_window:
                        html += """<td class="px-6 py-4"><a href='""" + hydrated_url + """' target="_blank" class="px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800 whitespace-nowrap">""" + hydrated_label + """</button></td>"""
                    else:
                        html += """<td class="px-6 py-4"><a href='""" + hydrated_url + """' class="px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800 whitespace-nowrap">""" + hydrated_label + """</button></td>"""
                else:
                    html += '''<td class="px-6 py-4">''' + format_input(row[column]) + "</td>"

        html += "</tr>"
        i += 1

    html += "</tbody>"

    html += "</table>"

    html += "</div>"

    html += "</div>"

    self.components.append(HtmlComponent(html))
    return self


"""
A function that formats input into a human-readable string:
Input: An arbitrary type that could be string, integer, floating point, a numpy object, a pandas datetime, or something else
Output: String

Dates should be formatted using ISO-8601. Numbers below 10 should include 2 decimal places. Numbers between 10 and 100 should have 1 decimal place. Numbers between 100 and 1000 should have 0 decimal places. Numbers between 1000 and 1000000 should have 0 decimal places and be formatted with a comma for the thousands separator. Numbers between 1000000 and 1000000000 should be formatted as X.Y million. Numbers above 1000000000 should be formatted as X.Y billion
"""


def format_input(input):
    is_int = "int" in type(input).__name__
    is_float = "float" in type(input).__name__

    if is_float or is_int:
        if input < 10:
            if is_float:
                return '{:.2f}'.format(input)
            else:
                return string_format_with_more(str(input), 10)
        elif input < 100:
            if is_float:
                return '{:.1f}'.format(input)
            else:
                return string_format_with_more(str(input), 10)
        elif input < 1000:
            return '{:.0f}'.format(input)
        elif input < 1000000:
            return '{:,.0f}'.format(input)
        elif input < 1000000000:
            return '{:.1f} million'.format(input / 1000000)
        # Check if it's nan
        elif input != input:
            return "N/A"
        else:
            return '{:.1f} billion'.format(input / 1000000000)
    elif callable(getattr(input, "isoformat", None)):
        iso = input.isoformat()

        if "T00:00:00" in iso:
            return iso[0:10]

        return iso
    else:
        formatted_string = str(input)
        return string_format_with_more(formatted_string, 100)

def string_format_with_more(text: str, max_length: int) -> str:
    if len(text) > max_length:
        return text[0:max_length] + f'... <button data-text-full="{text}" data-text-truncated="{text[0:max_length]}..." onclick="toggleMore(this)" class="text-blue-500">more</button>'
    else:
        return text

def __format_python_object_for_json(t):
    if callable(getattr(t, "isoformat", None)):
        iso = t.isoformat()

        if "T00:00:00" in iso:
            return iso[0:10]

        return iso
    
    return t

def advanced_add_emgithub(self, url):
    quoted_url = quote(url)

    emgithub = '''
    <script src="https://emgithub.com/embed-v2.js?target=''' + quoted_url +  '''&style=vs2015&type=code&showBorder=on&showLineNumbers=on&showFileMeta=on&showFullPath=on&showCopy=on&fetchFromJsDelivr=on"></script>
    '''

    self.components.append(HtmlComponent(emgithub))

def __format_column_header(x: str) -> str:
    # Insert undersore between camel case
    x = re.sub(r'(?<=[a-z])(?=[A-Z])', '_', x)

    # Capitalize the beginning of each word
    x = x.title()

    # Replace underscores with spaces
    x = x.replace('_', ' ')

    return x

def __get_action_buttons_to_add(action_buttons):
    action_buttons_to_add = []

    for action_button in action_buttons:
        # If the label is not of the form "{key}", then add it to the list of action buttons to add
        if action_button.label[0] != '{' or action_button.label[-1] != '}' and '{' not in action_button.label[1:-1]:        
            action_buttons_to_add.append(action_button)

    return action_buttons_to_add

def __find_key_in_action_buttons(key: str, action_buttons):
    for action_button in action_buttons:
        if action_button.label == '{' + key + '}' :
            return action_button

    return None

def __replace_text_with_button(record: dict, action_buttons) -> dict:
    new_record = {}

    for key in record.keys():
        action_button = __find_key_in_action_buttons(key, action_buttons)

        value = __format_python_object_for_json(record[key])

        if action_button is not None:
            hydrated_label = action_button.label.format(**record)
            hydrated_url = action_button.url.format(**record)
            if action_button.open_in_new_window:
                new_record[key] = """<a href='""" + hydrated_url + """' target="_blank" class="px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800 whitespace-nowrap mr-1">""" + hydrated_label + """</button>"""
            else:
                new_record[key] = """<a href='""" + hydrated_url + """' class="px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800 whitespace-nowrap mr-1">""" + hydrated_label + """</button>"""
        else:
            new_record[key] = value
    
    action_buttons_html = ''

    action_buttons_to_add = __get_action_buttons_to_add(action_buttons)
    for button_to_add in action_buttons_to_add:
        hydrated_label = button_to_add.label.format(**record)
        hydrated_url = button_to_add.url.format(**record)

        if button_to_add.open_in_new_window:
            action_buttons_html += """<a href='""" + hydrated_url + """' target="_blank" class="px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800 whitespace-nowrap">""" + hydrated_label + """</button>"""
        else:
            action_buttons_html += """<a href='""" + hydrated_url + """' class="px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800 whitespace-nowrap">""" + hydrated_label + """</button>"""

    new_record['Actions'] = action_buttons_html

    return new_record

def advanced_add_datagrid(page, dataframe, action_buttons):
    if action_buttons is None:
        action_buttons = []

    cols = list(map(lambda x: {'headerName': __format_column_header(x), 'field': x} , dataframe.columns.to_list()))

    if len(__get_action_buttons_to_add(action_buttons)) > 0:
        cols.append({'headerName': 'Actions', 'field': 'Actions'})

    datagridHtml = '''
        <script>
        var columnDefsasdf = {columns};
        '''.format(columns = cols)

    records = list (map(lambda x: __replace_text_with_button(x, action_buttons=action_buttons), dataframe.to_dict(orient='records')))

    datagridHtml += '''
    columnDefsasdf.forEach( (x) => { x.cellRenderer = function(params) { return params.value ? params.value : '' } } )
    '''

    datagridHtml += '''
        var rowDataasdf = {records};
        '''.format(records = json.dumps( records ) )

    datagridHtml += '''
        var gridOptionsasdf = {
            columnDefs: columnDefsasdf,
            rowData: rowDataasdf,
            defaultColDef: {
                sortable: true,
                filter: true,
                resizable: true,
                floatingFilter: true,
                autoHeight: true,
                wrapText: true,
                autoSizePadding: 10,
                cellStyle: {
                    'white-space': 'normal'
                },
            },
            pagination: true
        };
        document.addEventListener('DOMContentLoaded', function() {
            var gridDivasdf = document.querySelector('#divid_aggrid_asdf');
            new agGrid.Grid(gridDivasdf, gridOptionsasdf);
        });

        function expand(e) {
            e.parentElement.children[1].style.height = 'calc( 100vh )';
            e.scrollIntoView();
        }
        </script>
        <div>
            <button onclick="expand(this)" class="mb-4 px-3 py-2 text-xs font-medium text-center text-white bg-blue-700 rounded-lg hover:bg-blue-800 focus:ring-4 focus:outline-none focus:ring-blue-300 dark:bg-blue-600 dark:hover:bg-blue-700 dark:focus:ring-blue-800">Expand</button>
            <div id="divid_aggrid_asdf" style="height: 500px; max-height: calc( 100vh - 60px ); " class="data-grid ag-theme-alpine-dark "></div>
        </div>
    '''

    page.components.append(HtmlComponent(datagridHtml))

